import pandas as pd
import numpy as np

from sklearn import preprocessing as sklearn_preprocessing

from .commons import get_shape_diff


class ZscoreOutlierHandler:
    @staticmethod
    def convert_zscore(df: pd.DataFrame, cols: list = None):
        scaled = df.select_dtypes(include="number").astype(np.float64)
        if scaled.shape[1] == 0:
            print("No numeric columns to convert")
            return df
        ss = sklearn_preprocessing.StandardScaler()
        if cols is None:
            return pd.DataFrame(
                ss.fit_transform(scaled), columns=scaled.columns, index=scaled.index
            )
        else:
            scaled.loc[:, cols] = ss.fit_transform(scaled[cols])
            return pd.DataFrame(scaled, columns=scaled.columns, index=scaled.index)

    @staticmethod
    @get_shape_diff()
    def remove_outliers(
        df: pd.DataFrame,
        cols: list = None,
        upper_bound: int = 3,
        return_original_df: bool = False,
    ) -> pd.DataFrame:
        scaled = ZscoreOutlierHandler.convert_zscore(df.copy(), cols)
        if return_original_df:
            res = df.copy()
        else:
            res = scaled
        if cols is None:
            return res.loc[(np.abs(scaled) <= upper_bound).all(axis=1)]
        else:
            return res.loc[(np.abs(scaled[cols]) <= upper_bound).all(axis=1)]



from matplotlib import pyplot as plt
import seaborn as sns

import numpy as np


def plot_hist(df, sample=True, n_cols=3, **histplot_kwargs):
    sample_threshold = 100000
    if sample:
        sample_number = min(sample_threshold, df.shape[0])
    _df = df.sample(sample_number)

    _df = _df.select_dtypes(include="number")

    n_rows = _df.shape[1] / n_cols
    n_rows = np.ceil(n_rows)
    n_rows = int(n_rows)

    fig, axes = plt.subplots(nrows=n_rows, ncols=n_cols, figsize=(20, 15))

    for i, column in enumerate(_df.columns):
        sns.histplot(
            _df[column], bins=50, ax=axes[i // n_cols, i % n_cols], **histplot_kwargs
        )
    plt.tight_layout()



from functools import wraps


def get_num_diff_rows(remove_kind="rows"):
    def _calc_diff(method):
        @wraps(method)
        def _calc(*args, **kw):
            old_df = args[0]
            result_df = method(*args, **kw)
            delta_units = result_df.shape[0] - old_df.shape[0]
            delta_units_perc = delta_units / old_df.shape[0]
            print(
                f"{method.__name__} >> {delta_units:,} ({delta_units_perc:,.2%}) {remove_kind} removed"
            )
            return result_df

        return _calc

    return _calc_diff


def get_shape_diff():
    def _calc_diff(method):
        @wraps(method)
        def _calc(*args, **kw):
            old_df = args[0]
            result_df = method(*args, **kw)
            delta_rows = result_df.shape[0] - old_df.shape[0]
            delta_rows_perc = delta_rows / old_df.shape[0]
            delta_columns = result_df.shape[1] - old_df.shape[1]
            delta_columns_perc = delta_columns / old_df.shape[1]
            print(
                f"{method.__name__} >> Delta shape: ({delta_rows:,}, {delta_columns:,}) | ({delta_rows_perc:,.1%}, {delta_columns_perc:,.1%})"
            )
            return result_df

        return _calc

    return _calc_diff



import warnings
import inspect

from sklearn.base import BaseEstimator, TransformerMixin

import sklearn
import pandas as pd
import scipy
import numpy as np


class PersistFeatureNames(BaseEstimator, TransformerMixin):
    def __init__(self, prefix=""):
        self.prefix = prefix

    def fit(self, X, y=None):
        if isinstance(X, pd.DataFrame):
            self.feature_names = X.add_prefix(self.prefix).columns
        else:
            self.feature_names = pd.DataFrame(data=X).add_prefix(self.prefix).columns

    def transform(self, X):
        if isinstance(X, pd.DataFrame):
            return X.add_prefix(self.prefix)
        else:
            return pd.DataFrame(data=X, columns=self.feature_names)

    def fit_transform(self, X, y=None):
        self.fit(X, y)
        return self.transform(X)

    def get_feature_names(self):
        return self.feature_names


def FeatureNamesAsPandasCols(model, name=None):
    class _AppendFeatureNamesAsPandasCols(BaseEstimator, TransformerMixin):
        def __init__(self, model, name):
            self.model = model
            if name is None:
                self.name = model.__class__.__name__
            else:
                self.name = name
            self.feature_names = None

        def fit(self, X, y=None):
            return self.model.fit(X, y)

        def transform(self, X):
            data = self.model.transform(X)
            self.feature_names = self.extract_feature_names(X)
            return pd.DataFrame(data=data, columns=self.feature_names)

        def fit_transform(self, X, y=None):
            self.fit(X, y)
            return self.transform(X)

        def get_feature_names(self):
            return self.feature_names

        def __repr__(self):
            return self.name

    class FeatureGeneratorWrapper(_AppendFeatureNamesAsPandasCols):
        def transform(self, X):
            data = self.model.transform(X)
            self.feature_names = self.extract_feature_names(X)
            if scipy.sparse.issparse(data):
                df = pd.DataFrame.sparse.from_spmatrix(
                    data=data, columns=self.feature_names
                )
            else:
                df = pd.DataFrame(data=data, columns=self.feature_names)
            return df

        def extract_feature_names(self, X):
            """Extracts the feature names from arbitrary sklearn models

            Args:
            model: The Sklearn model, transformer, clustering algorithm, etc. which we want to get named features for.
            name: The name of the current step in the pipeline we are at.

            Returns:
            The list of feature names.  If the model does not have named features it constructs feature names
            by appending an index to the provided name.
            """
            model = self.model
            name = self.name

            if hasattr(model, "get_feature_names"):
                fn = model.get_feature_names
                if len(inspect.getargspec(fn).args) == 1:
                    feature_names = fn()
                else:
                    feature_names = fn(X.columns)
                self.feature_names = [f"{name}__{feature}" for feature in feature_names]
                return self.feature_names
            elif hasattr(model, "n_clusters"):
                self.feature_names = [f"{name}__{x}" for x in range(model.n_clusters)]
                return self.feature_names
            elif hasattr(model, "n_components"):
                self.feature_names = [f"{name}__{x}" for x in range(model.n_components)]
                return self.feature_names
            elif hasattr(model, "components_"):
                n_components = model.components_.shape[0]
                self.feature_names = [f"{name}__{x}" for x in range(n_components)]
                return self.feature_names
            elif hasattr(model, "classes_"):
                self.feature_names = [f"{name}__{_class}" for _class in model.classes_]
                return self.feature_names
            else:
                self.feature_names = [name]
                return self.feature_names

    class TransformerWrapper(_AppendFeatureNamesAsPandasCols):
        def extract_feature_names(self, X):
            if isinstance(X, pd.DataFrame):
                self.feature_names = X.add_suffix("__" + self.name).columns
            else:
                self.feature_names = (
                    pd.DataFrame(data=X).add_suffix("__" + self.name).columns
                )
            return self.feature_names

    class FeatureSelectorWrapper(_AppendFeatureNamesAsPandasCols):
        def extract_feature_names(self, X):
            # https://stackoverflow.com/questions/35376293/extracting-selected-feature-names-from-scikit-pipeline
            columns = self.model.transform([X.columns]).flatten()
            self.feature_names = [f"{col}__{self.name}" for col in columns]
            return self.feature_names

    if model.__class__.__name__ in ["SelectFromModel"]:
        return FeatureSelectorWrapper(model, name)
    if (
        hasattr(model, "get_feature_names")
        or hasattr(model, "n_clusters")
        or hasattr(model, "n_components")
        or hasattr(model, "kneighbors")  # KNN has attribute classes_ after fit
    ):
        return FeatureGeneratorWrapper(model, name)
    else:
        return TransformerWrapper(model, name)


def get_feature_names(column_transformer):
    """Get feature names from all transformers.
    Source: https://johaupt.github.io/scikit-learn/tutorial/python/data%20processing/ml%20pipeline/model%20interpretation/columnTransformer_feature_names.html

    Returns
    -------
    feature_names : list of strings
        Names of the features produced by transform.
    """
    # Remove the internal helper function
    # check_is_fitted(column_transformer)

    # Turn loopkup into function for better handling with pipeline later
    def get_names(trans):
        # >> Original get_feature_names() method
        if trans == "drop" or (hasattr(column, "__len__") and not len(column)):
            return []
        if trans == "passthrough":
            if hasattr(column_transformer, "_df_columns"):
                if (not isinstance(column, slice)) and all(
                    isinstance(col, str) for col in column
                ):
                    return column
                else:
                    return column_transformer._df_columns[column]
            else:
                indices = np.arange(column_transformer._n_features)
                return ["x%d" % i for i in indices[column]]
        if not hasattr(trans, "get_feature_names"):
            # >>> Change: Return input column names if no method avaiable
            # Turn error into a warning
            warnings.warn(
                "Transformer %s (type %s) does not "
                "provide get_feature_names. "
                "Will return input column names if available"
                % (str(name), type(trans).__name__)
            )
            # For transformers without a get_features_names method, use the input
            # names to the column transformer
            if column is None:
                return []
            else:
                return [name + "__" + f for f in column]

        return [name + "__" + f for f in trans.get_feature_names()]

    ### Start of processing
    feature_names = []

    # Allow transformers to be pipelines. Pipeline steps are named differently, so preprocessing is needed
    if type(column_transformer) == sklearn.pipeline.Pipeline:
        l_transformers = [
            (name, trans, None, None)
            for step, name, trans in column_transformer._iter()
        ]
    else:
        # For column transformers, follow the original method
        l_transformers = list(column_transformer._iter(fitted=True))

    for name, trans, column, _ in l_transformers:
        if type(trans) == sklearn.pipeline.Pipeline:
            # Recursive call on pipeline
            _names = get_feature_names(trans)
            # if pipeline has no transformer that returns names
            if len(_names) == 0:
                _names = [name + "__" + f for f in column]
            feature_names.extend(_names)
        else:
            feature_names.extend(get_names(trans))

    return feature_names



from functools import wraps
import contextlib

import numpy as np
import pandas as pd
import pandas.io.formats.format as pf

from IPython.display import display

from .outlier import ZscoreOutlierHandler


def describe_df(df, cardinality_ubound=10):
    @contextlib.contextmanager
    def float_and_int_formatting():
        orig_float_format = pd.options.display.float_format
        orig_int_format = pf.IntArrayFormatter

        pd.options.display.float_format = "{:0,.2f}".format

        class IntArrayFormatter(pf.GenericArrayFormatter):
            def _format_strings(self):
                formatter = self.formatter or "{:,d}".format
                fmt_values = [formatter(x) for x in self.values]
                return fmt_values

        pf.IntArrayFormatter = IntArrayFormatter
        yield
        pd.options.display.float_format = orig_float_format
        pf.IntArrayFormatter = orig_int_format

    # with pd.option_context('display.float_format', '{:,.2f}'.format).:
    with float_and_int_formatting():
        description_df = df.describe(include="all").T
        display(description_df.fillna(""))
    cat_description_df = df.select_dtypes(exclude="number")
    if (
        cat_description_df.shape[1] > 0
        and (description_df["unique"] <= cardinality_ubound).sum() > 0
    ):
        print("Categorical features distribution:")
        for c in cat_description_df:
            if df[c].nunique() <= cardinality_ubound:
                print(
                    c,
                    [
                        (k, v)
                        for k, v in df[c]
                        .value_counts(normalize=True, dropna=False)
                        .round(2)
                        .to_dict()
                        .items()
                    ],
                    sep=" ",
                )
    return


def summarize(df, count_missing=True, count_outliers=True, zscore_upperbound=3):
    display(df.head())
    display(df.shape)
    summary_df = pd.DataFrame(
        index=df.columns,
        data=np.nan,
        columns=["missing_count", "missing_perc", "outlier_count", "outlier_perc"],
    )
    summary_df["dtype"] = df.dtypes.apply(str)
    if count_missing:
        summary_df = summary_df.assign(
            missing_count=df.isnull().sum(axis=0),
            missing_perc=lambda s_df: (s_df["missing_count"] / df.shape[0]),
        )

    if count_outliers and df.select_dtypes(include="number").shape[1] > 0:
        scaled = ZscoreOutlierHandler.convert_zscore(df.select_dtypes(include="number"))
        summary_df = summary_df.assign(
            outlier_count=scaled.where(np.abs(scaled) > zscore_upperbound).count(
                axis=0
            ),
            outlier_perc=lambda df: (
                df["outlier_count"] / scaled.notnull().sum(axis=0)
            ),
        )

        summary_df = summary_df.assign(
            zero_count=(df.select_dtypes(include="number") == 0).sum(axis=0),
            zero_perc=lambda _df: (
                _df["zero_count"] / df.select_dtypes(include="number").shape[0]
            ),
            negative_count=(df.select_dtypes(include="number") < 0).sum(axis=0),
            negative_perc=lambda _df: (
                _df["negative_count"] / df.select_dtypes(include="number").shape[0]
            ),
        )
    print("\nSummary:")
    display(
        summary_df.replace(np.nan, 0)
        .loc[lambda df: ~((df["missing_count"] == 0) & (df["outlier_count"] == 0))]
        .sort_values(by=["missing_perc", "outlier_perc"], ascending=[False, False])
        .style.format(
            formatter=dict(
                zip(
                    [col for col in summary_df.columns if col.endswith("_perc")],
                    [
                        r"{:.2%}".format
                        for x in range(sum(summary_df.columns.str.endswith("_perc")))
                    ],
                )
            )
        )
        .format(
            formatter=dict(
                zip(
                    [col for col in summary_df.columns if col.endswith("_count")],
                    [
                        r"{:,.0f}".format
                        for x in range(sum(summary_df.columns.str.endswith("_count")))
                    ],
                )
            )
        )
        .background_gradient(axis=0)
    )

    if df.select_dtypes(include="number").shape[1] > 0:
        tmp = df.describe(include="number").T
        tmp.index.name = "Numeric features"
        print()
        display(tmp.style.format(r"{:,.1f}").background_gradient(axis=0))

    if df.select_dtypes(exclude="number").shape[1] > 0:
        tmp = df.describe(exclude="number").T
        tmp.index.name = "Categorical features"
        print()
        display(
            tmp.style.format(
                r"{:,.0f}", subset=[col for col in tmp.columns if col not in ["top"]]
            ).background_gradient(
                axis=0, subset=[col for col in tmp.columns if col not in ["top"]]
            )
        )
        for c in df.select_dtypes(exclude="number"):
            if df[c].nunique() <= 10:
                print(f"Categorical feature {c} distribution:")
                print(
                    c,
                    [
                        (k, v)
                        for k, v in df[c]
                        .value_counts(normalize=True, dropna=False)
                        .round(2)
                        .to_dict()
                        .items()
                    ],
                    sep=" ",
                )


def summarize_df(count_missing=True, count_outliers=True, zscore_upperbound=3):
    def _summarize_df(method):
        @wraps(method)
        def _summarize(*args, **kw):
            df = method(*args, **kw)
            if (
                kw.__contains__("activate_summarize_df_deco")
                and not kw["activate_summarize_df_deco"]
            ):
                return df
            summarize(
                df,
                count_missing=count_missing,
                count_outliers=count_outliers,
                zscore_upperbound=zscore_upperbound,
            )
            return df

        return summarize

    return _summarize_df



import time
from functools import wraps


def log_time(get_time=False, printer=print):
    def _log_time(method):
        @wraps(method)
        def timed(*args, **kw):
            ts = time.time()
            result = method(*args, **kw)
            te = time.time()
            walltime = te - ts
            printer("{} runtime: {:.0f}s".format(method.__name__, (te - ts)))
            if get_time:
                return result, walltime
            else:
                return result

        return timed

    return _log_time
